# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest tool file."""
import argparse
import enum
import logging
import sys

from manifesttool import __version__
from manifesttool.mtool.actions.create import CreateAction
from manifesttool.mtool.actions.parse import ParseAction
from manifesttool.mtool.actions.public_key import PublicKeyAction
from manifesttool.mtool.actions.schema import PrintSchemaAction
from manifesttool.mtool.asn1.v1 import ManifestAsnCodecV1
from manifesttool.mtool.asn1.v3 import ManifestAsnCodecV3

logger = logging.getLogger("manifest-tool")


class Actions(enum.Enum):
    """Actions class."""

    CREATE = "create"
    CREATE_V1 = "create-v1"
    PARSE = "parse"
    SCHEMA = "schema"
    PUB_KEY = "public-key"


def get_parser():
    """Get argument parser."""
    parser = argparse.ArgumentParser(
        description="Tool for creating, signing and verifying "
        "manifest files for running Pelion Device "
        "management update campaigns.",
        add_help=False,
    )

    parser.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )
    parser.add_argument(
        "--version",
        action="version",
        version="Manifest-Tool version {}".format(__version__),
        help="Show program's version number and exit.",
    )

    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Print error logs only."
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print exception info upon exiting.",
    )

    actions_parser = parser.add_subparsers(title="Commands", dest="action")
    actions_parser.required = True

    create_parser = actions_parser.add_parser(
        Actions.CREATE.value,
        help="Create a manifest.",
        description="Create a manifest.",
        add_help=False,
    )
    CreateAction.register_parser_args(
        create_parser, ManifestAsnCodecV3.get_name()
    )

    create_parser = actions_parser.add_parser(
        Actions.CREATE_V1.value,
        help="Create a V1 schema manifest.",
        description="Create a V1 schema manifest.",
        add_help=False,
    )
    CreateAction.register_parser_args(
        create_parser, ManifestAsnCodecV1.get_name()
    )

    verify_parser = actions_parser.add_parser(
        Actions.PARSE.value,
        help="Parse and verify a manifest against the input "
        "validation schema.",
        description="Parse and verify a manifest against the input "
        "validation schema.",
        add_help=False,
    )
    ParseAction.register_parser_args(verify_parser)

    schema_parser = actions_parser.add_parser(
        Actions.SCHEMA.value,
        help="Print the input validation schema.",
        description="Print the input validation schema.",
        add_help=False,
    )
    PrintSchemaAction.register_parser_args(schema_parser)

    public_key_parser = actions_parser.add_parser(
        Actions.PUB_KEY.value,
        help="Create a public key file containing a key in "
        "uncompressed point format.",
        description="Create a public key file containing a key in "
        "uncompressed point format.",
        add_help=False,
    )
    PublicKeyAction.register_parser_args(public_key_parser)

    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    """Entry point of the manifest tool."""
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format="%(asctime)s %(levelname)s %(message)s",
        level=logging.ERROR if args.quiet else logging.DEBUG,
    )

    if args.debug:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

    try:
        action = Actions(args.action)
        if action == Actions.PARSE:
            ParseAction.entry_point(args)
        elif action == Actions.CREATE:
            CreateAction.entry_point(args, ManifestAsnCodecV3)
        elif action == Actions.CREATE_V1:
            CreateAction.entry_point(args, ManifestAsnCodecV1)
        elif action == Actions.SCHEMA:
            PrintSchemaAction.entry_point(args)
        elif action == Actions.PUB_KEY:
            PublicKeyAction.entry_point(args)
        else:
            # will never get here
            raise AssertionError("Invalid action")
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(str(ex), exc_info=args.debug)
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
# Copyright 2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest-dev-tool create command."""
import argparse
import logging
import time
from pathlib import Path
from typing import Type

import yaml

from manifesttool.dev_tool import defaults
from manifesttool.common.common_helpers import get_argument_path
from manifesttool.common.common_helpers import get_non_negative_int_argument
from manifesttool.mtool.actions import semantic_version_arg_factory
from manifesttool.mtool.actions import semver_as_tuple_arg_factory
from manifesttool.mtool.actions.create import CreateAction
from manifesttool.mtool.asn1 import ManifestAsnCodecBase
from manifesttool.mtool.payload_format import PayloadFormat

logger = logging.getLogger("manifest-dev-tool-create")


def register_parser(
    parser: argparse.ArgumentParser,
    schema_version: str,
    is_update_parser: bool = False,
):
    """Register the parser."""
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")

    if not is_update_parser:
        required.add_argument(
            "-u",
            "--payload-url",
            help="Address from which the device downloads "
            "the candidate payload.",
            required=True,
        )

    required.add_argument(
        "-p",
        "--payload-path",
        help="Local path to the candidate payload file.",
        type=get_argument_path,
        required=True,
    )

    if not is_update_parser:
        required.add_argument(
            "-o",
            "--output",
            help="Output manifest filename.",
            type=argparse.FileType("wb"),
            required=True,
        )

    if schema_version == "v1":

        version_group = optional.add_mutually_exclusive_group()
        version_group.add_argument(
            "-v",
            "--fw-version",
            type=get_non_negative_int_argument,
            help="Version number (integer) of the candidate image. "
            "Default: current epoch time.",
            default=int(time.time()),
        )
        version_group.add_argument(
            "--fw-migrate-ver",
            type=semver_as_tuple_arg_factory,
            help="Version number of the candidate image in "
            "SemVer format. NOTE: Use to upgrade from "
            "v1 manifest schema to a later schema.",
        )

        optional.add_argument(
            "-r",
            "--priority",
            type=get_non_negative_int_argument,
            help="Set update priority >=0.",
        )
    else:  # v3
        if is_update_parser:
            optional.add_argument(
                "-e",
                "--encrypt-payload",
                action="store_true",
                help="Encrypt the payload the device downloads.",
            )
        else:
            optional.add_argument(
                "--encrypted-digest",
                type=str,
                help="Encrypted payload digest (32-byte HEX string)."
                " Use only if the candidate payload is encrypted.",
            )
            optional.add_argument(
                "--encrypted-size",
                type=get_non_negative_int_argument,
                help="Encrypted payload size."
                " Use only if the candidate payload is encrypted.",
            )
        optional.add_argument(
            "-v",
            "--fw-version",
            type=semantic_version_arg_factory,
            help="Version number of the candidate image in SemVer format. "
            "Min. 0.0.1, max 999.999.999. "
            "Must be bigger than the version currently "
            "in the device(s).",
        )
        optional.add_argument(
            "--component-name",
            default="MAIN",
            metavar="NAME",
            help="The name of the component to be updated. "
            "Must correspond to an existing component "
            "name on target devices. [Default: MAIN].",
        )
        if is_update_parser:
            optional.add_argument(
                "-u",
                "--use-short-url",
                action="store_true",
                help="Use a short candidate payload URL in the manifest. "
                "Note: the device must be configured to use CoAP.",
            )
        optional.add_argument(
            "-m",
            "--sign-image",
            action="store_true",
            help="Sign image. Use only when the bootloader on a device "
            "expects a signed FW image.",
        )
        optional.add_argument(
            "-r",
            "--priority",
            type=get_non_negative_int_argument,
            help="Set update priority >=0. [Default: 0].",
            default=0,
        )
        optional.add_argument(
            "--combined-image",
            action="store_true",
            help="Use combined package prepared by manifest-package-tool",
        )

    optional.add_argument(
        "-d",
        "--vendor-data",
        help="Path to a vendor custom data file - to be passed to "
        "the target devices.",
        type=get_argument_path,
    )

    optional.add_argument(
        "--cache-dir",
        help="Tool's cache directory. "
        'Must match the directory used by the "init" command.',
        type=Path,
        default=defaults.BASE_PATH,
    )

    optional.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )


def set_payload_format(delta_meta_file, combined_package, encrypted_digest):
    """Set the payload format."""
    if delta_meta_file.is_file():
        if encrypted_digest:
            raise AssertionError("Unexpected combination (Delta+Encrypted)")
        payload_format = PayloadFormat.PATCH
    elif combined_package is True:
        if encrypted_digest:
            payload_format = PayloadFormat.ENCRYPTED_COMBINED
        else:
            payload_format = PayloadFormat.COMBINED
    else:
        if encrypted_digest:
            payload_format = PayloadFormat.ENCRYPTED_RAW
        else:
            payload_format = PayloadFormat.RAW

    return payload_format


def create_dev_manifest(
    dev_cfg: dict,
    manifest_version: Type[ManifestAsnCodecBase],
    vendor_data_path: Path,
    payload_path: Path,
    payload_url: str,
    encrypted_digest: str,
    encrypted_size: int,
    priority: int,
    fw_version: str,
    sign_image: bool,
    component: str,
    combined_package: bool,
):
    """Create developer manifest."""
    payload_file = payload_path
    delta_meta_file = payload_file.with_suffix(".yaml")

    payload_format = set_payload_format(
        delta_meta_file, combined_package, encrypted_digest
    )

    input_cfg = {
        "vendor": {"vendor-id": dev_cfg["vendor-id"]},
        "device": {"class-id": dev_cfg["class-id"]},
        "payload": {
            "url": payload_url,
            "file-path": payload_path.as_posix(),
            "format": payload_format.value,
        },
    }

    if encrypted_digest:
        input_cfg["payload"]["encrypted"] = {
            "digest": encrypted_digest,
            "size": encrypted_size,
        }

    if priority is not None:
        input_cfg["priority"] = priority

    if manifest_version.get_name() != "v1":
        input_cfg["sign-image"] = sign_image
        input_cfg["component"] = component

    if vendor_data_path:
        input_cfg["vendor"]["custom-data-path"] = vendor_data_path.as_posix()

    signing_key = None
    if "signing-tool" in dev_cfg and dev_cfg["signing-tool"] is not None:
        input_cfg["signing-tool"] = dev_cfg["signing-tool"]
        signing_key = dev_cfg["key_file"]
    else:
        key_file = Path(dev_cfg["key_file"])
        if not key_file.is_file():
            raise AssertionError("{} not found".format(key_file.as_posix()))
        signing_key = key_file.read_bytes()

    logger.debug("input_cfg=\n%s", yaml.dump(input_cfg))

    manifest_bin = CreateAction.do_create(
        signing_key=signing_key,
        input_cfg=input_cfg,
        fw_version=fw_version,
        update_certificate=Path(dev_cfg["certificate"]),
        asn1_codec_class=manifest_version,
    )
    logger.info(
        "Created manifest in %s schema for %s update campaign",
        manifest_version.get_name(),
        "delta" if payload_format == PayloadFormat.PATCH else "full",
    )
    return manifest_bin


def bump_version(sem_ver: str):
    """Bump version."""
    nibbles = sem_ver.split(".")
    for i in range(2, -1, -1):
        nibble = int(nibbles[i]) + 1
        if nibble <= 999:
            nibbles[i] = nibble
            break
        nibbles[i] = 0
    return semantic_version_arg_factory("{}.{}.{}".format(*nibbles))


def load_cfg_and_get_fw_ver(
    args, manifest_version: Type[ManifestAsnCodecBase]
):
    """Load config file and get FW version."""
    if not args.cache_dir.is_dir():
        raise AssertionError(
            "Tool cache directory is missing. "
            'Execute "init" command to create it.'
        )

    component_name = getattr(args, "component_name", "MAIN")
    fw_migrate_ver = getattr(args, "fw_migrate_ver", None)
    cache_fw_version_file = args.cache_dir / defaults.UPDATE_VERSION
    cached_versions = {}
    fw_sem_ver = None

    # load dev_cfg
    with (args.cache_dir / defaults.DEV_CFG).open("rt") as fh:
        dev_cfg = yaml.safe_load(fh)

    # load cached_versions
    if "v1" not in manifest_version.get_name() or fw_migrate_ver:
        if cache_fw_version_file.is_file():
            with cache_fw_version_file.open("rt") as fh:
                cached_versions = yaml.safe_load(fh)

    # set fw_version
    if "v1" not in manifest_version.get_name():
        fw_version = (
            args.fw_version
            if args.fw_version
            else bump_version(cached_versions.get(component_name, "0.0.1"))
        )
        fw_sem_ver = fw_version
        logger.info("FW version: %s", fw_version)
    elif fw_migrate_ver:
        fw_version, fw_sem_ver = args.fw_migrate_ver
        logger.info("FW version: %d (%s)", fw_version, fw_sem_ver)
    else:
        fw_version = args.fw_version
        logger.info("FW version: %d", fw_version)

    # save cached_versions
    if fw_sem_ver:
        cached_versions[component_name] = fw_sem_ver
        with cache_fw_version_file.open("wt") as fh:
            yaml.dump(cached_versions, fh)

    return dev_cfg, fw_version


def entry_point(
    args,
    parser: argparse.ArgumentParser,
    manifest_version: Type[ManifestAsnCodecBase],
):
    """Entry point of the create command."""
    encrypted_digest = getattr(args, "encrypted_digest", None)
    encrypted_size = getattr(args, "encrypted_size", None)

    if encrypted_digest or encrypted_size:
        if not encrypted_digest or not encrypted_size:
            parser.error(
                "require --encrypted-digest and --encrypted-size "
                "or none of those."
            )

    dev_cfg, fw_version = load_cfg_and_get_fw_ver(args, manifest_version)

    manifest_bin = create_dev_manifest(
        dev_cfg=dev_cfg,
        manifest_version=manifest_version,
        vendor_data_path=args.vendor_data,
        payload_path=args.payload_path,
        payload_url=args.payload_url,
        encrypted_digest=encrypted_digest,
        encrypted_size=encrypted_size,
        priority=args.priority,
        fw_version=fw_version,
        sign_image=getattr(args, "sign_image", False),
        component=getattr(args, "component_name", None),
        combined_package=getattr(args, "combined_image", False),
    )

    with args.output as fh:
        fh.write(manifest_bin)



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Package tool module."""
import argparse
import logging
import sys
import enum

from manifesttool.package_tool.actions.create import CreateAction
from manifesttool.package_tool.actions.parse import ParseAction
from manifesttool import __version__

logger = logging.getLogger("manifest-package-tool")


class Actions(enum.Enum):
    """Actions class."""

    CREATE = "create"
    PARSE = "parse"


def get_parser():
    """Get parser."""
    parser = argparse.ArgumentParser(
        description="Tool for creating and verifying combined package files "
        "used for full update campaigns.",
        add_help=False,
    )
    required = parser.add_argument_group("required arguments")

    required.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )

    actions_parser = parser.add_subparsers(title="Commands", dest="action")
    actions_parser.required = True

    create_parser = actions_parser.add_parser(
        Actions.CREATE.value,
        help="Create a package.",
        description="Create a package.",
        add_help=False,
    )

    CreateAction.register_parser_args(create_parser)

    verify_parser = actions_parser.add_parser(
        Actions.PARSE.value,
        help="Parse and verify a package against the input "
        "validation schema.",
        description="Parse and verify a package against the input "
        "validation schema.",
        add_help=False,
    )
    ParseAction.register_parser_args(verify_parser)
    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    """Entry point to the package tool."""
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format="%(asctime)s %(levelname)s %(message)s",
        level=logging.DEBUG,
    )
    try:
        action = Actions(args.action)
        if action == Actions.PARSE:
            ParseAction.entry_point(args)
        elif action == Actions.CREATE:
            CreateAction.entry_point(args)
        else:
            # will never get here
            raise AssertionError("Invalid action")
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(
            str(ex),
            # exc_info=args.debug
        )
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Package tool create action."""
import argparse
import logging
from pathlib import Path
import jsonschema
import yaml
from manifesttool.common.common_helpers import get_argument_path
from manifesttool.package_tool.asn1.package_encoder import DescriptorAsnCodec
from manifesttool.package_tool.package_format.tar_package import (
    PackageFormatTar,
)

PTOOL_PATH = Path(__file__).resolve().parent.parent
logger = logging.getLogger("manifest-package-tool-create")


def _get_alignment_int_argument(value: str) -> int:
    """
    Construct non-negative integer value for an argument.

    :param value: input string
    :return: integer value
    """
    int_value = None
    try:
        int_value = int(value)
    except ValueError:
        pass
    if int_value is None or int_value < 1:
        raise argparse.ArgumentTypeError(
            '"{}" is an invalid alignment value, should be >0'.format(value)
        )
    return int_value


class CreateAction:
    """CreateAction class."""

    @staticmethod
    def register_parser_args(parser: argparse.ArgumentParser):
        """Register parser arguments."""
        required = parser.add_argument_group("required arguments")
        optional = parser.add_argument_group("optional arguments")

        required.add_argument(
            "-c",
            "--config",
            help="Path to the package tool configuration file.",
            metavar="YAML",
            type=get_argument_path,
            required=True,
        )

        required.add_argument(
            "-f",
            "--format",
            help="Package format type.",
            choices=["tar"],  # Add bin when implemented
            default="tar",
        )

        required.add_argument(
            "-a",
            "--image-alignment-size",
            help="Candidate storage read size, \
                used for image alignment. Relevant for embedded devices.",
            default=1,
            type=_get_alignment_int_argument,
        )

        required.add_argument(
            "-o",
            "--output",
            help="Output package filename.",
            type=Path,
            required=True,
        )

        optional.add_argument(
            "-h",
            "--help",
            action="help",
            help="Show this help message and exit.",
        )

    @staticmethod
    def do_create(input_cfg: dict, output_file, format_type):
        """Create method."""
        # validate input against input-schema.json
        schema_path = PTOOL_PATH / "package-schema.json"
        with schema_path.open("rb") as fh:
            input_schema = yaml.safe_load(fh)
            jsonschema.validate(input_cfg, input_schema)

        # create asn1 descriptor
        codec = DescriptorAsnCodec()
        asn1der = codec.encode_package_descriptor(input_cfg)

        if "tar" in format_type:
            package_format = PackageFormatTar()
        else:
            logger.error("Package tool supports only tar format")
            raise NotImplementedError

        package_format.create_package(output_file, input_cfg, asn1der)

    @classmethod
    def entry_point(cls, args: argparse.Namespace) -> None:
        """Entry point to create action."""
        output_file = args.output

        with open(args.config, "rb") as config_fh:
            input_cfg = yaml.safe_load(config_fh)

        cls.do_create(input_cfg, output_file, args.format)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest tool V3 encoder."""
import logging
from collections import OrderedDict
from typing import Optional

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.native.encoder import encode as native_encoder

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1.manifest_codec import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v3 import manifest_schema_v3 as manifest_schema


class ManifestAsnCodecV3(ManifestAsnCodecBase):
    """ManifestAsnCodecV3 class."""

    VERSION = "v3"

    logger = logging.getLogger("v3-manifest-codec")

    def __init__(self, dom=None):
        """Init method."""
        super().__init__()
        if dom:
            self.dom = dom
        else:
            self.dom = manifest_schema.Manifest()

    @classmethod
    def get_name(cls) -> str:
        """Get name."""
        return cls.VERSION

    def set_payload_version(self, version: str):
        """Set payload version."""
        assert isinstance(version, str)
        self.dom["payload-version"] = version

    def set_update_priority(self, priority: int):
        """Set update priority."""
        self.dom["update-priority"] = priority

    def set_vendor_id(self, vendor_id: bytes):
        """Set vendor ID."""
        self.dom["vendor-id"] = vendor_id

    def set_vendor_data(self, data: bytes):
        """Set vendor data."""
        self.dom["vendor-data"] = data

    def set_class_id(self, class_id: bytes):
        """Set class ID."""
        self.dom["class-id"] = class_id

    def set_device_id(self, device_id: bytes):
        """Set device ID."""
        self.dom["device-id"] = device_id

    def set_payload_fingerprint(self, digest: bytes, size: int):
        """Set payload fingerprint."""
        self.dom["payload-digest"] = digest
        self.dom["payload-size"] = size

    def set_payload_uri(self, uri: str):
        """Set payload URI."""
        self.dom["payload-uri"] = uri

    def set_payload_format(self, payload_format: str):
        """Set payload format."""
        self.dom["payload-format"] = payload_format

    def set_payload_metadata(
        self,
        installed_digest: bytes,
        installed_size: int,
        precursor_digest: bytes,
    ):
        """Set payload metadata."""
        delta_meta_dom = manifest_schema.PayloadMetadata()
        delta_meta_dom["installed-size"] = installed_size
        delta_meta_dom["installed-digest"] = installed_digest
        if precursor_digest:
            delta_meta_dom["precursor-digest"] = precursor_digest
        self.dom["payload-metadata"] = delta_meta_dom

    @classmethod
    def decode(
        cls, data: bytes, verification_key: Optional[ec.EllipticCurvePublicKey]
    ) -> OrderedDict:
        """Decode."""
        signed_resource_dom = der_decoder.decode(
            data, asn1Spec=manifest_schema.SignedResource()
        )[0]
        if not verification_key:
            return native_encoder(signed_resource_dom)

        codec = ManifestAsnCodecV3(dom=signed_resource_dom["manifest"])
        signed_data = codec.get_signed_data()
        der_signature = bytes(signed_resource_dom["signature"])
        signature = ecdsa_helper.signature_raw_to_der(der_signature)

        try:
            ecdsa_helper.ecdsa_verify(
                public_key=verification_key,
                signed_data=signed_data,
                signature=signature,
            )
            cls.logger.info("Manifest Signature verified!")
        except InvalidSignature as ex:
            raise AssertionError(
                "Manifest Signature verification failed"
            ) from ex

        raw_image_signature = bytes(codec.dom["installed-signature"])
        if raw_image_signature:
            der_image_signature = ecdsa_helper.signature_raw_to_der(
                raw_image_signature
            )
            digest = bytes(codec.dom["payload-digest"])
            if str(codec.dom["payload-format"]) == "arm-patch-stream":
                digest = bytes(
                    codec.dom["payload-metadata"]["installed-digest"]
                )
            try:
                ecdsa_helper.ecdsa_verify_prehashed(
                    public_key=verification_key,
                    digest=digest,
                    signature=der_image_signature,
                )
                cls.logger.info("Image Signature verified!")
            except InvalidSignature as ex:
                raise AssertionError(
                    "Image Signature verification failed"
                ) from ex

        return native_encoder(signed_resource_dom)

    def get_signed_data(self) -> bytes:
        """Get signed data."""
        # sha_content = utils.sha_hash(enc_data)

        return der_encoder.encode(self.dom)

    def get_der_signed_resource(self, signature: bytes) -> bytes:
        """Get DER signed resource."""
        resource_dom = manifest_schema.SignedResource()
        resource_dom["manifest-version"] = self.VERSION
        resource_dom["manifest"] = self.dom
        resource_dom["signature"] = signature
        return der_encoder.encode(resource_dom)

    def set_update_certificate(self, cert_data: bytes):
        """Set update certificate."""
        raise NotImplementedError

    def set_image_signature(self, signature: bytes):
        """Set image signature."""
        self.dom["installed-signature"] = signature

    def set_component_name(self, component: str):
        """Set component name."""
        self.dom["component-name"] = component



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Parser action."""
import argparse
import binascii
import collections
import logging

import yaml
from pyasn1.error import PyAsn1Error

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1 import ManifestVersion


def bytes_representer(dumper: yaml.Dumper, obj):
    """Bytes representer."""
    try:
        return dumper.represent_str(obj.decode("utf-8"))
    except UnicodeError:
        return dumper.represent_str(binascii.hexlify(obj).decode("utf-8"))


def ordered_dict_representer(dumper, data):
    """Ordered dictionary representer."""
    return dumper.represent_dict(data.items())


yaml.Dumper.add_representer(collections.OrderedDict, ordered_dict_representer)
yaml.Dumper.add_representer(bytes, bytes_representer)


class ParseAction:
    """ParseAction class."""

    logger = logging.getLogger("manifest-tool-parse")

    @staticmethod
    def register_parser_args(parser):
        """Register parser arguments."""
        required = parser.add_argument_group("required arguments")
        optional = parser.add_argument_group("optional arguments")

        required.add_argument(
            "manifest",
            help="Path to the manifest file.",
            type=argparse.FileType("rb"),
        )

        key_or_cert_group = optional.add_mutually_exclusive_group()
        key_or_cert_group.add_argument(
            "-c",
            "--certificate",
            help="Path to a certificate file to validate "
            "the manifest signature.",
            type=argparse.FileType("rb"),
        )
        key_or_cert_group.add_argument(
            "-p",
            "--public-key",
            help="Path to a public key file, "
            "containing a key in uncompressed point format, "
            "to validate the manifest signature.",
            type=argparse.FileType("rb"),
        )
        key_or_cert_group.add_argument(
            "-k",
            "--private-key",
            help="Path to a private key PEM file "
            "to validate the manifest signature.",
            type=argparse.FileType("rb"),
        )

        optional.add_argument(
            "-h",
            "--help",
            action="help",
            help="Show this help message and exit.",
        )

    # pylint: disable=too-many-branches
    @classmethod
    def do_parse(
        cls,
        manifest_data: bytes,
        certificate_data: bytes,
        public_key_data: bytes,
        private_key_data: bytes,
    ):
        """Parse method."""
        dom = None

        public_key = None
        if private_key_data:
            public_key = ecdsa_helper.public_key_from_private(private_key_data)
        elif certificate_data:
            try:
                public_key = ecdsa_helper.public_key_from_certificate(
                    certificate_data
                )
            except ValueError as ex:
                raise AssertionError("Malformed certificate") from ex
        elif public_key_data:
            public_key = ecdsa_helper.public_key_from_bytes(public_key_data)

        for codec_class in ManifestVersion.list_codecs():
            try:
                dom = codec_class.decode(manifest_data, public_key)
                break
            except PyAsn1Error:
                continue
        if not dom:
            raise AssertionError("Malformed manifest")

        logging.info(
            "\n----- Manifest dump start -----\n"
            "%s----- Manifest dump end -----",
            yaml.dump(dom, default_flow_style=False),
        )

    @classmethod
    def entry_point(cls, args):
        """Entry point method."""
        cert_data = args.certificate.read() if args.certificate else None
        public_key = args.public_key.read() if args.public_key else None
        private_key = args.private_key.read() if args.private_key else None
        manifest_data = args.manifest.read()

        cls.do_parse(
            manifest_data=manifest_data,
            certificate_data=cert_data,
            public_key_data=public_key,
            private_key_data=private_key,
        )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
# Auto-generated by asn1ate v.0.6.0 from manifest_v3.asn
# (last modified on 2021-06-22 12:17:55.466130)
"""Manifest tool V3 schema."""
from pyasn1.type import (
    univ,
    char,
    namedtype,
    namedval,
    tag,
    constraint,
    useful,
)


class PayloadMetadata(univ.Sequence):
    """PayloadMetadata class."""

    pass


PayloadMetadata.componentType = namedtype.NamedTypes(
    namedtype.NamedType("installed-size", univ.Integer()),
    namedtype.NamedType("installed-digest", univ.OctetString()),
    namedtype.OptionalNamedType("precursor-digest", univ.OctetString()),
)


class Manifest(univ.Sequence):
    """Manifest class."""

    pass


Manifest.componentType = namedtype.NamedTypes(
    namedtype.NamedType("vendor-id", univ.OctetString()),
    namedtype.NamedType("class-id", univ.OctetString()),
    namedtype.NamedType("update-priority", univ.Integer()),
    namedtype.NamedType("component-name", char.UTF8String()),
    namedtype.NamedType("payload-version", char.UTF8String()),
    namedtype.NamedType("payload-digest", univ.OctetString()),
    namedtype.NamedType("payload-size", univ.Integer()),
    namedtype.NamedType("payload-uri", char.UTF8String()),
    namedtype.NamedType(
        "payload-format",
        univ.Enumerated(
            namedValues=namedval.NamedValues(
                ("raw-binary", 1),
                ("arm-patch-stream", 5),
                ("combined", 6),
                ("encrypted-raw", 257),
                ("encrypted-combined", 262),
            )
        ),
    ),
    namedtype.NamedType("installed-signature", univ.OctetString()),
    namedtype.OptionalNamedType("payload-metadata", PayloadMetadata()),
    namedtype.OptionalNamedType("vendor-data", univ.OctetString()),
)


class SignedResource(univ.Sequence):
    """SignedResource class."""

    pass


SignedResource.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "manifest-version",
        univ.Enumerated(namedValues=namedval.NamedValues(("v3", 3))),
    ),
    namedtype.NamedType("manifest", Manifest()),
    namedtype.NamedType("signature", univ.OctetString()),
)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Delta tool."""
import argparse
import base64
import hashlib
import logging
import re
import sys
from mmap import mmap, ACCESS_READ
from pathlib import Path

import yaml
from manifesttool.common.common_helpers import get_argument_path
from manifesttool import __version__
from manifesttool import armbsdiff

logger = logging.getLogger("manifest-delta-tool")


def _block_size_factory(value):
    """Check the block size parameter."""
    prospective = None
    try:
        prospective = int(value)
    except ValueError:
        pass
    if not prospective or prospective < 128:
        raise argparse.ArgumentTypeError(
            "{} is invalid - must be at least 128".format(value)
        )
    return prospective


def digest_file(file_path: Path):
    """Perform SHA256 on the file."""
    read_block_size = 65536
    with file_path.open("rb") as fh:
        hash_ctx = hashlib.sha256()
        buf = fh.read(read_block_size)
        while buf:
            hash_ctx.update(buf)
            buf = fh.read(read_block_size)
        file_len = fh.tell()
    return base64.b64encode(hash_ctx.digest()), file_len


def size_check(new_size, delta_size, threshold):
    """
    Assert new fw image is smaller than delta file.

    :param new_size: full new FW image size
    :param delta_size: delta file size
    :param threshold: size difference threshold for aborting the generation.
                      in case evaluates to False - size check will be aborted.
    """
    diff = 100 * float(delta_size) / float(new_size)

    if threshold and diff >= threshold:
        raise AssertionError(
            "Difference with delta image and update image is more than "
            "{} percent! Percentage is: {:.2f}".format(threshold, diff)
        )


def get_version_string_from_bin_file(fname: Path):
    """Get version from the binary file."""
    with fname.open("rb") as fh:
        with mmap(fh.fileno(), 0, access=ACCESS_READ) as my_mmap:
            match = re.search(b"(PELION/BSDIFF\\d{3})", my_mmap)
            if not match:
                raise AssertionError("Version details not found")
            return match.group(0).decode("utf-8")


def check_bsdiff_bspatch_versions(original_image_path: Path):
    """Check if bsdfif and bspatch versions match."""
    bsdiff_version = armbsdiff.get_version()
    logger.info("Current tool version %s", bsdiff_version)
    bspatch_version = get_version_string_from_bin_file(original_image_path)

    if bsdiff_version != bspatch_version:
        logger.error(
            "Bspatch version in {} is incomatible with this version "
            "of delta-tool."
        )

        logger.error(
            "Original image version is: %s",
            bspatch_version.decode("utf-8").split("/")[1],
        )

        logger.error(
            "Current bsdiff version is: %s",
            bsdiff_version.decode("utf-8").split("/")[1],
        )

        raise AssertionError("bsdiff/bspatch version mismatch")


def generate_delta(
    orig_fw: Path, new_fw: Path, output_delta_file: Path, block_size, threshold
):
    """Generate the delta."""
    check_bsdiff_bspatch_versions(orig_fw)

    original_digest, _ = digest_file(orig_fw)
    new_digest, new_size = digest_file(new_fw)

    if original_digest == new_digest:
        logger.warning(
            "New and old file are binary same. This will generate "
            "delta that will not change the original image. "
            "This is probably a mistake"
        )

    try:
        armbsdiff.generate(
            orig_fw.as_posix(),
            new_fw.as_posix(),
            output_delta_file.as_posix(),
            block_size,
        )

        _, delta_size = digest_file(output_delta_file)

        size_check(new_size, delta_size, threshold)
    except AssertionError:
        if output_delta_file.is_file():
            output_delta_file.unlink()
        raise

    delta_cfg_file = output_delta_file.with_suffix(".yaml")
    config = {
        "installed-digest": new_digest,
        "installed-size": new_size,
        "precursor-digest": original_digest,
    }
    with delta_cfg_file.open("wt") as fh:
        yaml.dump(config, fh)


def get_parser():
    """Get argument parser."""
    parser = argparse.ArgumentParser(
        description="Generate delta patch files to be used for "
        "delta update campaigns.",
        add_help=False,
    )
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")

    required.add_argument(
        "-c",
        "--current-fw",
        type=get_argument_path,
        help="Path to the currently installed firmware image, without "
        "headers, for delta update calculation.",
        required=True,
    )
    required.add_argument(
        "-n",
        "--new-fw",
        type=get_argument_path,
        help="Path to the candidate image.",
        required=True,
    )

    required.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Output delta patch filename. "
        "NOTE: The delta tool generates an additional "
        "configuration file with the same name but with a '.yaml' "
        "extension. The manifest tool needs both files, "
        "but only this output file must "
        "be uploaded to Pelion storage.",
        required=True,
    )

    optional.add_argument(
        "-b",
        "--block-size",
        type=_block_size_factory,
        help="Compression block size algorithm. "
        "A greater size provides better "
        "compression, but consumes more memory on the device. "
        "Default is 512 bytes. Minimum is 128 bytes. "
        "NOTE: This value MUST be aligned with the "
        "network (COAP/HTTP) buffer size used for download.",
        default=512,
    )

    size_group = optional.add_mutually_exclusive_group()

    # must be first in a group for default value to be set properly
    size_group.add_argument(
        "-t",
        "--threshold",
        type=int,
        choices=range(30, 100),
        metavar="[30-100]",
        default=60,
        help="The ratio of the delta patch size compared to the "
        "candidate image size above which to raise an exception. "
        "Default is 60.",
    )
    size_group.add_argument(
        "--skip-size-check",
        action="store_false",
        dest="threshold",
        help="Skip threshold validations.",
    )

    optional.add_argument(
        "--debug",
        action="store_true",
        help="Print exception info upon exiting.",
    )

    optional.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )
    optional.add_argument(
        "--version",
        action="version",
        version="Manifest-Tool version {}".format(__version__),
        help="Show program's version number and exit.",
    )

    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    """Entry point of the delta tool."""
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format="%(asctime)s %(levelname)s %(message)s",
        level=logging.DEBUG,
    )

    if args.debug:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

    try:
        generate_delta(
            orig_fw=args.current_fw,
            new_fw=args.new_fw,
            output_delta_file=args.output,
            block_size=args.block_size,
            threshold=args.threshold,
        )
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(str(ex), exc_info=args.debug)
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
# Copyright 2022-2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest-dev-tool init command."""
import argparse
import datetime
import logging
import time
import uuid
from pathlib import Path
import shutil
from typing import Optional

import yaml
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from manifesttool import __version__
from manifesttool.dev_tool import defaults
from manifesttool.mtool import ecdsa_helper

SCRIPT_DIR = Path(__file__).resolve().parent

logger = logging.getLogger("manifest-dev-tool-init")


def uuid_factory(value: str) -> uuid.UUID:
    """
    Construct UUID from a string command line argument.

    :param value: input string
    :return: UUID
    """
    try:
        return uuid.UUID(hex=value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(
            '"{}" is a malformed hexadecimal UUID string'.format(value)
        ) from exc


def register_parser(parser: argparse.ArgumentParser):
    """Register the parser."""
    optional = parser.add_argument_group("optional arguments")

    optional.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )

    optional.add_argument(
        "--cache-dir",
        help="Tool's cache directory for preserving the state between "
        '"init", "create" and "update" command invocations. '
        "[Default: {}]".format(defaults.BASE_PATH),
        type=Path,
        default=defaults.BASE_PATH,
    )

    optional.add_argument(
        "-g",
        "--generated-resource",
        help="Generated update resource C filename. "
        "[Default: {}]".format(defaults.UPDATE_RESOURCE_C),
        type=Path,
        default=defaults.UPDATE_RESOURCE_C,
    )

    optional.add_argument(
        "--vendor-id",
        help="Set custom vendor UUID. [Default: random UUID]",
        type=uuid_factory,
        default=uuid.uuid4(),
    )

    optional.add_argument(
        "--class-id",
        help="Set custom class UUID. [Default: random UUID]",
        type=uuid_factory,
        default=uuid.uuid4(),
    )

    optional.add_argument(
        "--update-certificate",
        help="""
        Path to the update certificate file.
        The parameter must come together with `--update-certificate` argument
        """,
        type=Path,
    )

    optional.add_argument(
        "--key",
        help="""
        A private key file that sings the manifest.
        Could be a path to the PEM format private key file.
        It could also be an identifier for a private key,
        if `signing-tool` parameter is supplied.
        The `--key` parameter must come together with
        `--update-certificate` argument.
        """,
        type=Path,
    )

    optional.add_argument(
        "-s",
        "--signing-tool",
        help="""
             An external tool that signs the manifest.
             This allows signing with existing infrastructure.
             The arguments to the tool are:
             <digest algorithm> <key identifier> <input file> <output file>.
             Only SHA256 is currently supported as <digest algorithm>.
             The parameter must come together with `--key`
             and `--update-certificate` arguments.
             The `--key` will be used as <key identifier>
             """,
        type=Path,
    )

    service = parser.add_argument_group(
        "optional arguments (Izuma Device Management service configuration)"
    )

    service.add_argument(
        "-p",
        "--gw-preset",
        help="The preset name defined in {}, "
        " which specifies a URL and "
        "access key.".format(defaults.IZUMA_GW_PATH),
        choices=defaults.IZUMA_GW.keys() if defaults.IZUMA_GW else [],
    )
    service.add_argument(
        "-a",
        "--access-key",
        "--api-key",
        help="Access key for accessing a Izuma Device Management API.",
    )
    service.add_argument(
        "-u",
        "--api-url",
        help="Izuma Device Management API URL. "
        "[Default: {}]".format(defaults.API_GW),
        default=defaults.API_GW,
    )


def chunks(arr, num_of_elements):
    """Split the array into chunks of number of elements."""
    for i in range(0, len(arr), num_of_elements):
        yield arr[i : i + num_of_elements]


def format_rows(arr):
    """Format the chunks as hexadecimal values in a string."""
    for _slice in chunks(arr, 8):
        yield ", ".join(["0x{:02X}".format(x) for x in _slice])


def pretty_print(arr):
    """Create list from the format rows."""
    return ",\n    ".join(list(format_rows(arr)))


def generate_update_default_resources_c(
    c_source: Path,
    vendor_id: uuid.UUID,
    class_id: uuid.UUID,
    cert_file: Path,
):
    """
    Generate update resources C source file for developer convenience.

    :param c_source: generated C source file
    :param vendor_id: vendor UUID
    :param class_id: class UUID
    :param cert_file: update certificate file
    """
    vendor_id_str = pretty_print(vendor_id.bytes)
    class_id_str = pretty_print(class_id.bytes)

    cert_data = cert_file.read_bytes()

    cert_str = pretty_print(cert_data)

    template = (SCRIPT_DIR / "code_template.txt").read_text()

    public_key = ecdsa_helper.public_key_from_certificate(cert_data)
    public_key_bytes = ecdsa_helper.public_key_to_bytes(public_key)
    update_public_key_str = pretty_print(public_key_bytes)

    c_source.write_text(
        template.format(
            vendor_id=vendor_id_str,
            class_id=class_id_str,
            cert=cert_str,
            timestamp=time.strftime("%Y-%m-%d %H:%M:%S %Z"),
            tool="manifest-dev-tool init",
            version=__version__,
            update_pub_key=update_public_key_str,
        )
    )
    logger.info("generated update source %s", c_source)


def import_credentials(
    origin_cert_file: Path,
    dest_cert_file: Path,
    origin_key_file: Optional[Path],
    dest_key_file: Optional[Path],
):
    """
    Import developer credentials.

    :param origin_cert_file - path to .der public key certificate file
    :param dest_cert_file - destination file path
    :param origin_key_file - If provided, path to .pem signing key file
    :param dest_key_file - If provided, destination file path

    """
    logger.info("importing dev-credentials")

    if origin_key_file is not None:
        if origin_key_file != dest_key_file:
            shutil.copy(origin_key_file, dest_key_file)
            logger.info("imported %s to %s", origin_key_file, dest_key_file)

    if origin_cert_file != dest_cert_file:
        shutil.copy(origin_cert_file, dest_cert_file)
        logger.info("imported %s to %s", origin_cert_file, dest_cert_file)

    logger.info("dev-credentials - imported")


def generate_credentials(
    key_file: Path, cert_file: Path, cred_valid_time: int
):
    """
    Generate developer credentials.

    :param key_file - path to .pem signing key file
    :param cert_file - path to .der public key certificate file
    :param cred_valid_time - x.509 certificate validity period
    """
    try:
        logger.info("generating dev-credentials")
        key = ec.generate_private_key(ec.SECP256R1(), default_backend())

        # For a self-signed certificate the
        # subject and issuer are always the same.
        subject_list = [
            x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, "localhost")
        ]

        subject = issuer = x509.Name(subject_list)

        subject_key = key.public_key().public_bytes(
            serialization.Encoding.DER,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        hash_ctx = hashes.Hash(hashes.SHA256(), backend=default_backend())
        hash_ctx.update(subject_key)
        key_digest = hash_ctx.finalize()

        subject_key_identifier = key_digest[
            : 160 // 8
        ]  # Use RFC7093, Method 1

        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                datetime.datetime.utcnow()
                + datetime.timedelta(days=cred_valid_time)
            )
            .add_extension(
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=False,
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    key_cert_sign=False,
                    crl_sign=False,
                    encipher_only=False,
                    decipher_only=False,
                ),
                critical=False,
            )
            .add_extension(
                x509.SubjectAlternativeName([x509.DNSName("localhost")]),
                critical=False,
            )
            .add_extension(
                x509.ExtendedKeyUsage(
                    [x509.oid.ExtendedKeyUsageOID.CODE_SIGNING]
                ),
                critical=False,
            )
            .add_extension(
                x509.SubjectKeyIdentifier(subject_key_identifier),
                critical=False,
                # Sign our certificate with our private key
            )
            .sign(key, hashes.SHA256(), default_backend())
        )

        key_file.parent.mkdir(parents=True, exist_ok=True)
        key_file.write_bytes(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
        logger.info("created %s", key_file)

        cert_file.parent.mkdir(parents=True, exist_ok=True)
        cert_file.write_bytes(cert.public_bytes(serialization.Encoding.DER))

        logger.info("created %s", cert_file)

        logger.info("dev-credentials - generated")

    except ValueError:
        logger.error("failed to generate dev-credentials")

        if key_file.is_file():
            key_file.unlink()
        if cert_file.is_file():
            cert_file.unlink()
        raise


def generate_developer_config(
    key,
    cert_file: Path,
    config: Path,
    vendor_id: uuid.UUID,
    class_id: uuid.UUID,
    signing_tool: Optional[Path] = None,
):
    """
    Generate developer config.

    :param key - can be a path to .pem signing key file or
    a key identifier in case `signing_tool` is configured
    :param cert_file - path to .der public key certificate file
    :param config - config file
    :param vendor_id - vendor ID
    :param class_id - class ID
    :param signing_tool - optional parameter, external signing tool
    """
    cfg_data = {
        "key_file": str(key),
        "certificate": cert_file.as_posix(),
        "class-id": class_id.bytes.hex(),
        "vendor-id": vendor_id.bytes.hex(),
    }

    if signing_tool is not None:
        cfg_data["signing-tool"] = signing_tool.as_posix()

    config.parent.mkdir(parents=True, exist_ok=True)
    with config.open("wt") as fh:
        yaml.dump(cfg_data, fh)

    logger.info("generated developer config file %s", config)
    logger.debug("input_cfg=\n%s", yaml.dump(cfg_data))


def generate_service_config(
    access_key: str, api_url: str, api_config_path: Path
):
    """
    Generate service config.

    :param access_key - access key to the Izuma service
    :param api_url - API URL of the Izuma service
    :param api_config_path - path to a config file
    """
    cfg = {}

    if access_key is None or api_url is None:
        return

    if api_config_path.is_file():
        with api_config_path.open("rt") as fh:
            cfg = yaml.safe_load(fh)
            # if generated with previous versions
            # delete api_key
            if "api_key" in cfg:
                cfg.pop("api_key")

    cfg["access_key"] = access_key
    cfg["host"] = api_url

    with api_config_path.open("wt") as fh:
        yaml.safe_dump(cfg, fh)

    logger.info("generated service config file %s", api_config_path)


def entry_point(args, parser: argparse.ArgumentParser):
    """Entry point."""
    cache_dir = args.cache_dir
    cache_dir.mkdir(parents=True, exist_ok=True)

    vendor_id = args.vendor_id
    class_id = args.class_id

    signing_tool = getattr(args, "signing_tool", None)
    key = getattr(args, "key", None)
    update_certificate = getattr(args, "update_certificate", None)

    if signing_tool and (not key or not update_certificate):
        parser.error(
            "--signing-tool requires also --key and --update-certificate."
        )

    if key or update_certificate:
        if not update_certificate or not key:
            parser.error(
                "require both --key and --update-certificate or none of those."
            )

        import_cert = update_certificate
        import_key = None
        dest_key_file = None
        if not signing_tool:
            # If signing tool isn't supplied the given key should be imported
            import_key = key
            dest_key_file = cache_dir / defaults.UPDATE_PRIVATE_KEY

        import_credentials(
            origin_cert_file=import_cert,
            dest_cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
            origin_key_file=import_key,
            dest_key_file=dest_key_file,
        )
    else:
        # If a key and a certificate aren't given, they will be generated
        key = cache_dir / defaults.UPDATE_PRIVATE_KEY
        generate_credentials(
            key_file=key,
            cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
            cred_valid_time=365 * 20,  # years
        )

    generate_developer_config(
        key=key,
        cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
        config=cache_dir / defaults.DEV_CFG,
        vendor_id=vendor_id,
        class_id=class_id,
        signing_tool=signing_tool,
    )

    generate_update_default_resources_c(
        c_source=args.generated_resource,
        vendor_id=vendor_id,
        class_id=class_id,
        cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
    )
    access_key = args.access_key
    if not access_key and hasattr(args, "gw_preset") and args.gw_preset:
        access_key = defaults.IZUMA_GW[args.gw_preset].get("access_key")
        if not access_key:
            access_key = defaults.IZUMA_GW[args.gw_preset].get("api_key")

    api_url = args.api_url
    if hasattr(args, "gw_preset") and args.gw_preset:
        api_url = defaults.IZUMA_GW[args.gw_preset].get("host")

    generate_service_config(
        access_key=access_key,
        api_url=api_url,
        api_config_path=cache_dir / defaults.CLOUD_CFG,
    )

    if not access_key:
        logger.warning(
            "Access key not provided, so assisted "
            "campaign management will not be available"
        )

    (cache_dir / defaults.DEV_README).write_text(
        'Files in this directory are autogenerated by "manifest-dev-tool init"'
        " tool for internal use only.\nDo not modify.\n"
    )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Schema action."""
from pathlib import Path

MTOOL_PATH = Path(__file__).resolve().parent.parent


class PrintSchemaAction:
    """PrintSchemaAction class."""

    @staticmethod
    def print_schema():
        """Print schema."""
        schema_file = MTOOL_PATH / "manifest-input-schema.json"
        print(schema_file.read_text())

    @staticmethod
    def register_parser_args(parser):
        """Register parser arguments."""
        optional = parser.add_argument_group("optional arguments")

        optional.add_argument(
            "-h",
            "--help",
            action="help",
            help="Show this help message and exit.",
        )

    @classmethod
    def entry_point(cls, _):
        """Entry point method."""
        cls.print_schema()



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
# Copyright 2022-2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Defaults for the dev-tool."""
from pathlib import Path

import yaml

BASE_PATH = Path(".manifest-dev-tool")
UPDATE_PUBLIC_KEY_CERT = "dev.cert.der"
UPDATE_PRIVATE_KEY = "dev.key.pem"
DEV_CFG = "dev.cfg.yaml"
UPDATE_RESOURCE_C = "update_default_resources.c"
CLOUD_CFG = "dev.cloud_cfg.yaml"
DEV_README = "README.txt"
API_GW = "https://api.us-east-1.mbedcloud.com"
UPDATE_VERSION = "update.version.yaml"

IZUMA_GW_PATH = Path.home() / ".izuma-dev-presets.yaml"
IZUMA_GW = None
if IZUMA_GW_PATH.is_file():
    with IZUMA_GW_PATH.open("rb") as fh:
        IZUMA_GW = yaml.safe_load(fh)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Create action."""
import argparse
import logging
import time
from pathlib import Path
from typing import Type

import tempfile
import subprocess
import os
import jsonschema
import yaml

from manifesttool.common.common_helpers import get_argument_path
from manifesttool.common.common_helpers import get_non_negative_int_argument
from manifesttool.mtool.actions import semantic_version_arg_factory
from manifesttool.mtool.actions import semver_as_tuple_arg_factory
from manifesttool.mtool.asn1 import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v1 import ManifestAsnCodecV1
from manifesttool.mtool import ecdsa_helper

MTOOL_PATH = Path(__file__).resolve().parent.parent

logger = logging.getLogger("manifest-tool-create")


def create_signature(
    input_cfg: dict,
    raw_signature: bool,
    der_manifest: bytes,
    signing_key,
) -> bytes:
    """Create signature."""
    if "signing-tool" in input_cfg:
        try:
            signing_tool = input_cfg["signing-tool"]
            signing_tool_path = os.path.join(os.getcwd(), signing_tool)
            digest_algo = "sha256"
            infile = None
            with tempfile.NamedTemporaryFile(delete=False) as f:
                infile = f.name
                f.write(der_manifest)
                f.flush()
                logger.debug("Temporary manifest file: %s", infile)
            outfile = None
            with tempfile.NamedTemporaryFile(delete=False) as f:
                outfile = f.name
                logger.debug("Temporary signature file: %s", outfile)

            cmd = [
                signing_tool_path,
                digest_algo,
                signing_key,
                infile,
                outfile,
            ]
            logger.info("Running %s to sign manifest.", (" ".join(cmd)))
            # pylint: disable=R1732
            process = subprocess.Popen(cmd)  # nosec
            process.wait()
            process.terminate()
        except Exception:  # pylint: disable=broad-except
            logger.critical("Signing tool failed.")
            raise
        with open(outfile, "rb") as f:
            signature = f.read()
    else:
        logger.debug("no external signing tool")
        signature = ecdsa_helper.ecdsa_sign(der_manifest, signing_key)
    if raw_signature:
        signature = ecdsa_helper.signature_der_to_raw(signature)

    return signature


class CreateAction:
    """CreateAction class."""

    @staticmethod
    def register_parser_args(
        parser: argparse.ArgumentParser, schema_version: str
    ):
        """Register parser arguments."""
        required = parser.add_argument_group("required arguments")
        optional = parser.add_argument_group("optional arguments")

        required.add_argument(
            "-c",
            "--config",
            help="Path to the manifest tool configuration file.",
            metavar="YAML",
            type=argparse.FileType("rb"),
            required=True,
        )

        required.add_argument(
            "-k",
            "--key",
            help="""
            A private key file that sings the manifest.
            Could be a path to the PEM format private key file.
            It could also be an identifier for a private key,
            if `signing-tool` value is supplied in the configuration file
            """,
            metavar="KEY",
            required=True,
        )

        if schema_version == "v1":
            version_group = optional.add_mutually_exclusive_group()
            version_group.add_argument(
                "-v",
                "--fw-version",
                type=get_non_negative_int_argument,
                help="Version number (integer) of the candidate image. "
                "Default: current epoch time.",
                default=int(time.time()),
            )
            version_group.add_argument(
                "--fw-migrate-ver",
                type=semver_as_tuple_arg_factory,
                help="Version number of the candidate image in "
                "SemVer format. NOTE: Use to upgrade from "
                "v1 manifest schema to a later schema.",
            )

            required.add_argument(
                "--update-certificate",
                type=get_argument_path,
                help="Path to the update certificate file.",
                required=True,
            )
        else:
            required.add_argument(
                "-v",
                "--fw-version",
                type=semantic_version_arg_factory,
                help="Version number of the candidate image in SemVer format. "
                "Min. 0.0.1, max 999.999.999. "
                "Must be bigger than the version currently "
                "in the device(s).",
                required=True,
            )

        required.add_argument(
            "-o",
            "--output",
            help="Output manifest filename.",
            type=argparse.FileType("wb"),
            required=True,
        )

        optional.add_argument(
            "-h",
            "--help",
            action="help",
            help="Show this help message and exit.",
        )

    @staticmethod
    def do_create(
        signing_key,
        input_cfg: dict,
        fw_version,
        update_certificate: Path,
        asn1_codec_class: Type[ManifestAsnCodecBase],
    ) -> bytes:
        """Create method."""
        assert fw_version is not None

        codec = asn1_codec_class()

        # validate input against manifest-input-schema.json
        schema_path = MTOOL_PATH / "manifest-input-schema.json"
        with schema_path.open("rb") as fh:
            input_schema = yaml.safe_load(fh)
            if isinstance(codec, ManifestAsnCodecV1):
                # priority field is optional for v1
                # delete it from required list
                input_schema["required"].remove("priority")
            jsonschema.validate(input_cfg, input_schema)

        raw_signature = True
        if isinstance(codec, ManifestAsnCodecV1):
            raw_signature = False
            cert_data = update_certificate.read_bytes()
            codec.set_update_certificate(cert_data)
        installed_digest = codec.process_input_config(fw_version, input_cfg)

        if input_cfg.get("sign-image"):
            if isinstance(codec, ManifestAsnCodecV1):
                raise AssertionError(
                    "sign-image is unexpected for manifest schema v1"
                )
            if "signing-tool" in input_cfg:
                raise AssertionError(
                    "sign-image option is not supported together \
                    with signing-tool option"
                )
            signature = ecdsa_helper.ecdsa_sign_prehashed(
                installed_digest, signing_key
            )
            codec.set_image_signature(
                ecdsa_helper.signature_der_to_raw(signature)
            )
        else:
            codec.set_image_signature(bytes())

        der_manifest = codec.get_signed_data()

        signature = create_signature(
            input_cfg, raw_signature, der_manifest, signing_key
        )
        manifest_bin = codec.get_der_signed_resource(signature)

        logger.info(
            "Attention: When updating Mbed OS devices,"
            " candidate features must match the device's"
            " bootloader features. Incompatibility may"
            " result in damaged devices."
        )

        return manifest_bin

    @classmethod
    def entry_point(
        cls, args: argparse.Namespace, asn1_codec: Type[ManifestAsnCodecBase]
    ) -> None:
        """Entry point method."""
        input_cfg = yaml.safe_load(args.config)

        if getattr(args, "fw_migrate_ver", None):
            fw_version = args.fw_migrate_ver[0]
        else:
            fw_version = args.fw_version

        if "signing-tool" in input_cfg:
            # If signing-tool option is supplied
            # The key argument is propagated as is, without any manipulations
            # It will be used, later on, together with the signing-tool
            signing_key = args.key
        else:
            key_file = Path(args.key)
            signing_key = key_file.read_bytes()

        manifest_bin = cls.do_create(
            signing_key=signing_key,
            input_cfg=input_cfg,
            fw_version=fw_version,
            update_certificate=getattr(args, "update_certificate", None),
            asn1_codec_class=asn1_codec,
        )
        with args.output as fh:
            fh.write(manifest_bin)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
# Copyright 2022-2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest-dev-tool update command."""
import argparse
import logging
import time
from pathlib import Path
from typing import Type

import requests

from manifesttool.dev_tool import defaults
from manifesttool.dev_tool.actions.create import load_cfg_and_get_fw_ver
from manifesttool.dev_tool.actions.create import (
    register_parser as register_create_parser,
)
from manifesttool.dev_tool.actions.create import create_dev_manifest
from manifesttool.dev_tool.izuma import izuma
from manifesttool.common.common_helpers import get_non_negative_int_argument
from manifesttool.mtool.asn1 import ManifestAsnCodecBase

logger = logging.getLogger("manifest-dev-tool-update")

# The update API has a maximum name length of 128, but this is not queryable.
MAX_NAME_LEN = 128


def register_parser(parser: argparse.ArgumentParser, schema_version: str):
    """Register the parser."""
    register_create_parser(parser, schema_version, True)

    pdm_group = parser.add_argument_group(
        "optional arguments (Tool's campaign management)"
    )

    pdm_group.add_argument(
        "-i",
        "--device-id",
        help="Device ID for targeting a specific device in "
        "update campaign filter.",
    )

    pdm_group.add_argument(
        "-s",
        "--start-campaign",
        action="store_true",
        help="Create and start update campaign automatically.",
    )
    pdm_group.add_argument(
        "-w",
        "--wait-for-completion",
        dest="wait",
        action="store_true",
        help="Create and start update campaign automatically, "
        "wait for it to finish, and clean up created resources.",
    )
    pdm_group.add_argument(
        "-t",
        "--timeout",
        type=get_non_negative_int_argument,
        help="Wait timeout in seconds. " "[Default: 0 - wait-forever]",
        default=0,
    )
    pdm_group.add_argument(
        "-n",
        "--no-cleanup",
        action="store_true",
        help="Skip cleaning up created resources. "
        "Only relevant if --wait-for-completion was provided.",
    )


def _manage_campaign(
    api: izuma.IzumaServiceApi,
    campaign_id: izuma.ID,
    end_time: int,
    do_wait: bool,
):
    starting = True
    old_state = ""
    try:
        api.campaign_start(campaign_id)
    except requests.HTTPError:
        logger.error("Failed to start campaign %s", campaign_id)
        raise
    try:
        while end_time == 0 or time.time() < end_time:
            try:
                curr_campaign = api.campaign_get(campaign_id)
            except requests.ConnectionError as ex:
                logger.debug("Connection error %s", ex)
                time.sleep(3)
                continue
            curr_state = curr_campaign["state"]
            curr_phase = curr_campaign["phase"]
            if starting:
                if api.campaign_is_not_started(curr_phase):
                    raise AssertionError(
                        "Campaign not started - check filter and campaign "
                        "state.\n"
                        "Reason: {}".format(curr_campaign["autostop_reason"])
                    )
                if api.campaign_is_active(curr_phase):
                    logger.info("Started Campaign ID: %s", campaign_id)
                    starting = False
                    if not do_wait:
                        return
            else:
                if not api.campaign_is_active(curr_phase):
                    logger.info(
                        "Campaign is finished in state: %s", curr_state
                    )
                    return
            if old_state != curr_state:
                logger.info("Campaign state: %s", curr_state)
                old_state = curr_state
            time.sleep(3)
        logger.error("Campaign timed out")
        raise AssertionError("Campaign timed out")
    except requests.HTTPError:
        logger.error("Failed to retrieve campaign state")
        raise


def _print_summary(summary: dict, summary_reasons: dict):
    logger.info("----------------------------")
    logger.info("    Campaign Summary ")
    logger.info("----------------------------")
    logger.info(
        " Successfully updated:   %d",
        summary.get("success") if summary.get("success") else 0,
    )
    logger.info(
        " Failed to update:       %d",
        summary.get("fail") if summary.get("fail") else 0,
    )
    logger.info(
        " Skipped:                %d",
        summary.get("skipped") if summary.get("skipped") else 0,
    )
    logger.info(
        " Pending:                %d",
        summary.get("info") if summary.get("info") else 0,
    )
    logger.info(" Total in this campaign: %d", sum(summary.values()))
    if summary_reasons.get("fail"):
        logger.warning("Reasons for failed updates:")
        for reasons in summary_reasons["fail"]:
            if reasons.get("description"):
                logger.warning(" %s", reasons.get("description"))
    if summary_reasons.get("skipped"):
        logger.warning("Reasons for skipped updates:")
        for reasons in summary_reasons["skipped"]:
            if reasons.get("description"):
                logger.warning(" %s", reasons.get("description"))


def _finalize(
    api: izuma.IzumaServiceApi,
    do_cleanup: bool,
    campaign_id: izuma.ID,
    manifest_id: izuma.ID,
    fw_image_id: izuma.ID,
    manifest_path: Path,
):
    summary = {}
    summary_reasons = {}
    failed_devices = []
    try:
        if campaign_id:
            # stop the campaign if it's still active
            api.campaign_stop(campaign_id)
            # get summary, failed devices and reasons
            statistics = api.campaign_statistics(campaign_id)
            summary = {s["id"]: s["count"] for s in statistics}
            if summary.get("fail") and summary.get("fail") > 0:
                devices_state = api.campaign_device_metadata(campaign_id)
                failed_devices = [
                    d["device_id"]
                    for d in devices_state
                    if d["deployment_state"] != "deployed"
                ]
                summary_reasons["fail"] = api.campaign_statistic_events(
                    campaign_id, "fail"
                )
            if summary.get("skipped") and summary.get("skipped") > 0:
                summary_reasons["skipped"] = api.campaign_statistic_events(
                    campaign_id, "skipped"
                )
        if do_cleanup:
            logger.info("Cleaning up resources...")
            if campaign_id:
                api.campaign_delete(campaign_id)
            if manifest_id:
                api.manifest_delete(manifest_id)
            if fw_image_id:
                api.fw_delete(fw_image_id)
            if manifest_path and manifest_path.is_file():
                manifest_path.unlink()
    except requests.HTTPError as ex:
        logger.error("Failed to finalize update campaign")
        logger.debug("Exception %s", ex, exc_info=True)

    # print campaign summary
    if summary:
        _print_summary(summary, summary_reasons)
        if failed_devices:
            # assert if not all devices were updated
            raise AssertionError(
                "Failed to update {} devices: {}".format(
                    len(failed_devices), ", ".join(failed_devices)
                )
            )


def update(
    payload_path: Path,
    dev_cfg: dict,
    manifest_version: Type[ManifestAsnCodecBase],
    priority: int,
    vendor_data: Path,
    device_id: str,
    do_wait: bool,
    do_start: bool,
    end_time: int,
    do_cleanup: bool,
    service_config: Path,
    fw_version: str,
    sign_image: bool,
    component: str,
    short_url: bool,
    encrypt_payload: bool,
    combined_package: bool,
):
    """Update command."""
    api = izuma.IzumaServiceApi(service_config)
    manifest_path = None
    fw_image_id = None
    manifest_id = None
    campaign_id = None
    encrypted_digest = None
    encrypted_size = None
    try:
        timestamp = time.strftime("%Y_%m_%d-%H_%M_%S")
        logger.info("Uploading FW image %s", payload_path.as_posix())

        fw_meta = api.fw_upload(
            fw_name="{}-{}".format(timestamp, payload_path.name),
            image=payload_path,
            encrypt=encrypt_payload,
        )

        fw_image_id = fw_meta["id"]
        payload_url = fw_meta["short_datafile" if short_url else "datafile"]

        if encrypt_payload:
            if (
                fw_meta.get("datafile_encryption", False)
                and fw_meta.get("encrypted_datafile_checksum", False)
                and fw_meta.get("encrypted_datafile_size", False)
            ):
                encrypted_digest = fw_meta["encrypted_datafile_checksum"]
                encrypted_size = fw_meta["encrypted_datafile_size"]
            else:
                raise AssertionError(
                    "Request to encrypt payload failed! "
                    "Device Management doesn't support this feature."
                )

        manifest_data = create_dev_manifest(
            dev_cfg=dev_cfg,
            manifest_version=manifest_version,
            vendor_data_path=vendor_data,
            payload_path=payload_path,
            payload_url=payload_url,
            encrypted_digest=encrypted_digest,
            encrypted_size=encrypted_size,
            priority=priority,
            fw_version=fw_version,
            sign_image=sign_image,
            component=component,
            combined_package=combined_package,
        )

        manifest_name = "manifest-{timestamp}-{filename}".format(
            filename=payload_path.name, timestamp=timestamp
        )
        manifest_path = payload_path.parent / manifest_name

        manifest_path.write_bytes(manifest_data)
        manifest_id = api.manifest_upload(
            name=manifest_name, manifest=manifest_path
        )

        campaign_name = "campaign-{timestamp}-{filename}".format(
            filename=payload_path.name, timestamp=timestamp
        )

        filters = [
            "device_class__eq={}".format(dev_cfg["class-id"]),
            "vendor_id__eq={}".format(dev_cfg["vendor-id"]),
        ]
        if device_id:
            filters.append("id__eq={}".format(device_id))
        filters_query = "&".join(filters)
        campaign_id = api.campaign_create(
            name=campaign_name,
            manifest_id=manifest_id,
            device_filter=filters_query,
        )

        if do_start:
            _manage_campaign(api, campaign_id, end_time, do_wait)

    except requests.HTTPError as ex:
        # log additional service info
        srv_message = ex.response.json().get("message", None)
        if srv_message:
            logger.error(srv_message)
        raise
    except KeyboardInterrupt:
        logger.error("Aborted by user...")
    finally:
        if do_wait:
            _finalize(
                api,
                do_cleanup,
                campaign_id,
                manifest_id,
                fw_image_id,
                manifest_path,
            )


def entry_point(args, manifest_version: Type[ManifestAsnCodecBase]):
    """Entry point of the update command."""
    dev_cfg, fw_version = load_cfg_and_get_fw_ver(args, manifest_version)

    encrypt_payload = hasattr(args, "encrypt_payload") and args.encrypt_payload

    update(
        payload_path=args.payload_path,
        dev_cfg=dev_cfg,
        manifest_version=manifest_version,
        priority=args.priority,
        vendor_data=args.vendor_data,
        device_id=args.device_id,
        do_start=args.start_campaign or args.wait,
        do_wait=args.wait,
        end_time=time.time() + args.timeout if args.timeout > 0 else 0,
        do_cleanup=not args.no_cleanup,
        service_config=args.cache_dir / defaults.CLOUD_CFG,
        fw_version=fw_version,
        sign_image=getattr(args, "sign_image", False),
        component=getattr(args, "component_name", None),
        short_url=hasattr(args, "use_short_url") and args.use_short_url,
        encrypt_payload=encrypt_payload,
        combined_package=getattr(args, "combined_image", False),
    )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
# Auto-generated by asn1ate v.0.6.0 from manifest-1.0.0
# (last modified on 2020-01-26 09:34:14.860967)
"""Manifest tool V1 schema."""
from pyasn1.type import (
    univ,
    char,
    namedtype,
    namedval,
    tag,
    constraint,
    useful,
)


class Bytes(univ.OctetString):
    """Bytes class."""

    pass


class Uri(char.UTF8String):
    """URI class."""

    pass


class CertificateReference(univ.Sequence):
    """CertificateReference class."""

    pass


CertificateReference.componentType = namedtype.NamedTypes(
    namedtype.NamedType("fingerprint", Bytes()),
    namedtype.NamedType("uri", Uri()),
)


class KeyTableEntry(univ.Sequence):
    """KeyTableEntry class."""

    pass


KeyTableEntry.componentType = namedtype.NamedTypes(
    namedtype.OptionalNamedType(
        "hash",
        univ.OctetString().subtype(
            implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
        ),
    ),
    namedtype.OptionalNamedType(
        "payloadKey",
        univ.OctetString().subtype(
            implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
        ),
    ),
)


class MacBlock(univ.Sequence):
    """MacBlock class."""

    pass


MacBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType("pskID", univ.OctetString()),
    namedtype.NamedType("keyTableVersion", univ.Integer()),
    namedtype.OptionalNamedType("keyTableIV", univ.OctetString()),
    namedtype.OptionalNamedType("keyTableRef", char.UTF8String()),
    namedtype.NamedType("keyTableIndexSize", univ.Integer()),
    namedtype.NamedType("keyTableRecordSize", univ.Integer()),
)


class ResourceReference(univ.Sequence):
    """ResourceReference class."""

    pass


ResourceReference.componentType = namedtype.NamedTypes(
    namedtype.NamedType("hash", univ.OctetString()),
    namedtype.OptionalNamedType("uri", Uri()),
    namedtype.NamedType("size", univ.Integer()),
)


class PayloadDescription(univ.Sequence):
    """PayloadDescription class."""

    pass


PayloadDescription.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "format",
        univ.Choice(
            componentType=namedtype.NamedTypes(
                namedtype.NamedType(
                    "enum",
                    univ.Enumerated(
                        namedValues=namedval.NamedValues(
                            ("undefined", 0),
                            ("raw-binary", 1),
                            ("cbor", 2),
                            ("hex-location-length-data", 3),
                            ("elf", 4),
                            ("bsdiff-stream", 5),
                        )
                    ),
                ),
                namedtype.NamedType("objectId", univ.ObjectIdentifier()),
            )
        ),
    ),
    namedtype.OptionalNamedType(
        "encryptionInfo",
        univ.Sequence(
            componentType=namedtype.NamedTypes(
                namedtype.NamedType("initVector", univ.OctetString()),
                namedtype.NamedType(
                    "id",
                    univ.Choice(
                        componentType=namedtype.NamedTypes(
                            namedtype.NamedType("key", univ.OctetString()),
                            namedtype.NamedType(
                                "certificate", CertificateReference()
                            ),
                        )
                    ),
                ),
                namedtype.OptionalNamedType(
                    "key",
                    univ.Choice(
                        componentType=namedtype.NamedTypes(
                            namedtype.NamedType("keyTable", Uri()),
                            namedtype.NamedType(
                                "cipherKey", univ.OctetString()
                            ),
                        )
                    ),
                ),
            )
        ),
    ),
    namedtype.NamedType("storageIdentifier", char.UTF8String()),
    namedtype.NamedType("reference", ResourceReference()),
    namedtype.OptionalNamedType("installedSize", univ.Integer()),
    namedtype.OptionalNamedType("installedDigest", univ.OctetString()),
    namedtype.OptionalNamedType("version", char.UTF8String()),
)


class ResourceAlias(univ.Sequence):
    """ResourceAlias class."""

    pass


ResourceAlias.componentType = namedtype.NamedTypes(
    namedtype.NamedType("hash", univ.OctetString()),
    namedtype.NamedType("uri", Uri()),
)


class UUID(univ.OctetString):
    """UUID class."""

    pass


class Manifest(univ.Sequence):
    """Manifest class."""

    pass


Manifest.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "manifestVersion",
        univ.Enumerated(namedValues=namedval.NamedValues(("v1", 1))),
    ),
    namedtype.OptionalNamedType("description", char.UTF8String()),
    namedtype.NamedType("timestamp", univ.Integer()),
    namedtype.NamedType("vendorId", UUID()),
    namedtype.NamedType("classId", UUID()),
    namedtype.NamedType("deviceId", UUID()),
    namedtype.NamedType("nonce", univ.OctetString()),
    namedtype.NamedType("vendorInfo", univ.OctetString()),
    namedtype.OptionalNamedType("precursorDigest", univ.OctetString()),
    namedtype.OptionalNamedType(
        "applyPeriod",
        univ.Sequence(
            componentType=namedtype.NamedTypes(
                namedtype.NamedType("validFrom", univ.Integer()),
                namedtype.NamedType("validTo", univ.Integer()),
            )
        ),
    ),
    namedtype.NamedType("applyImmediately", univ.Boolean()),
    namedtype.OptionalNamedType("priority", univ.Integer()),
    namedtype.NamedType(
        "encryptionMode",
        univ.Choice(
            componentType=namedtype.NamedTypes(
                namedtype.NamedType(
                    "enum",
                    univ.Enumerated(
                        namedValues=namedval.NamedValues(
                            ("invalid", 0),
                            ("aes-128-ctr-ecc-secp256r1-sha256", 1),
                            ("none-ecc-secp256r1-sha256", 2),
                            ("none-none-sha256", 3),
                            ("none-psk-aes-128-ccm-sha256", 4),
                            ("aes-128-ccm-psk-sha256", 5),
                        )
                    ),
                ),
                namedtype.NamedType("objectId", univ.ObjectIdentifier()),
            )
        ),
    ),
    namedtype.NamedType(
        "aliases", univ.SequenceOf(componentType=ResourceAlias())
    ),
    namedtype.NamedType(
        "dependencies", univ.SequenceOf(componentType=ResourceReference())
    ),
    namedtype.OptionalNamedType("payload", PayloadDescription()),
)


class Payload(univ.OctetString):
    """Payload class."""

    pass


class Resource(univ.Sequence):
    """Resource class."""

    pass


Resource.componentType = namedtype.NamedTypes(
    namedtype.OptionalNamedType("uri", Uri()),
    namedtype.NamedType(
        "resourceType",
        univ.Enumerated(
            namedValues=namedval.NamedValues(("manifest", 0), ("payload", 1))
        ),
    ),
    namedtype.NamedType(
        "resource",
        univ.Choice(
            componentType=namedtype.NamedTypes(
                namedtype.NamedType("manifest", Manifest()),
                namedtype.NamedType("payload", Payload()),
            )
        ),
    ),
)


class SignatureBlock(univ.Sequence):
    """SignatureBlock class."""

    pass


SignatureBlock.componentType = namedtype.NamedTypes(
    namedtype.NamedType("signature", univ.OctetString()),
    namedtype.NamedType(
        "certificates", univ.SequenceOf(componentType=CertificateReference())
    ),
)


class ResourceSignature(univ.Sequence):
    """ResourceSignature class."""

    pass


ResourceSignature.componentType = namedtype.NamedTypes(
    namedtype.NamedType("hash", univ.OctetString()),
    namedtype.NamedType(
        "signatures", univ.SequenceOf(componentType=SignatureBlock())
    ),
    namedtype.OptionalNamedType(
        "macs", univ.SequenceOf(componentType=MacBlock())
    ),
)


class SignedResource(univ.Sequence):
    """SignedResource class."""

    pass


SignedResource.componentType = namedtype.NamedTypes(
    namedtype.NamedType("resource", Resource()),
    namedtype.NamedType("signature", ResourceSignature()),
)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Payload format options."""
import enum


class PayloadFormat(enum.Enum):
    """PayloadFormat class."""

    RAW = "raw-binary"
    PATCH = "arm-patch-stream"
    COMBINED = "combined"
    ENCRYPTED_RAW = "encrypted-raw"
    ENCRYPTED_COMBINED = "encrypted-combined"



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Manifest tool V1 encoder."""
import hashlib
import logging
import time
from collections import OrderedDict
from typing import Optional

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.native.encoder import encode as native_encoder
from pyasn1.type.univ import SequenceOf

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1.manifest_codec import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v1 import manifest_schema_v1 as schema


class ManifestAsnCodecV1(ManifestAsnCodecBase):
    """ManifestAsnCodecV1 class."""

    VERSION = "v1"

    logger = logging.getLogger("v1-manifest-codec")

    def __init__(self, dom=None):
        """Init method."""
        # Call the __init__ method of the base class.
        super().__init__()
        if dom:
            self.dom = dom
        else:
            self.dom = schema.Resource()
            self.dom["resourceType"] = "manifest"
            self.dom["resource"]["manifest"] = schema.Manifest()
            self.dom["resource"]["manifest"]["manifestVersion"] = "v1"
            self.dom["resource"]["manifest"]["nonce"] = b""
            self.dom["resource"]["manifest"]["deviceId"] = b""
            self.dom["resource"]["manifest"]["vendorInfo"] = b""
            self.dom["resource"]["manifest"]["applyImmediately"] = True
            self.dom["resource"]["manifest"]["encryptionMode"][
                "enum"
            ] = "none-ecc-secp256r1-sha256"
            self.dom["resource"]["manifest"][
                "payload"
            ] = schema.PayloadDescription()
            self.dom["resource"]["manifest"]["payload"][
                "storageIdentifier"
            ] = "default"
            self.dom["resource"]["manifest"]["payload"][
                "reference"
            ] = schema.ResourceReference()
            self.fingerprint = None

    @classmethod
    def get_name(cls) -> str:
        """Get name."""
        return cls.VERSION

    def get_der_signed_resource(self, signature: bytes) -> bytes:
        """Get DER signed resource."""
        hash_ctx = hashlib.sha256()
        hash_ctx.update(self.get_signed_data())
        signed_digest = hash_ctx.digest()

        resource_signature_dom = schema.ResourceSignature()
        resource_signature_dom["hash"] = signed_digest
        signature_block = schema.SignatureBlock()
        signature_block["signature"] = signature
        signature_block["certificates"] = SequenceOf()
        certificate_reference = schema.CertificateReference()
        certificate_reference["fingerprint"] = self.fingerprint
        certificate_reference["uri"] = ""
        signature_block["certificates"].append(certificate_reference)
        resource_signature_dom["signatures"] = SequenceOf()
        resource_signature_dom["signatures"].append(signature_block)

        signed_resource_dom = schema.SignedResource()
        signed_resource_dom["resource"] = self.dom
        signed_resource_dom["signature"] = resource_signature_dom

        return der_encoder.encode(signed_resource_dom)

    def get_signed_data(self) -> bytes:
        """Get signed data."""
        return der_encoder.encode(self.dom)

    @classmethod
    def decode(
        cls, data: bytes, verification_key: Optional[ec.EllipticCurvePublicKey]
    ) -> OrderedDict:
        """Decode."""
        signed_resource_dom = der_decoder.decode(
            data, asn1Spec=schema.SignedResource()
        )[0]
        if not verification_key:
            return native_encoder(signed_resource_dom)

        codec = ManifestAsnCodecV1(dom=signed_resource_dom["resource"])
        signed_data = codec.get_signed_data()
        signature = bytes(
            signed_resource_dom["signature"]["signatures"][0]["signature"]
        )

        try:
            ecdsa_helper.ecdsa_verify(
                public_key=verification_key,
                signed_data=signed_data,
                signature=signature,
            )
            cls.logger.info("Manifest Signature verified!")
        except InvalidSignature as ex:
            raise AssertionError(
                "Manifest Signature verification failed"
            ) from ex
        return native_encoder(signed_resource_dom)

    def set_payload_version(self, version: str):
        """Set payload version."""
        if version:
            try:
                version_number = int(version)
            except ValueError as ex:
                raise AssertionError(
                    "invalid version {} - version must be "
                    "a valid positive integer".format(version)
                ) from ex
        else:
            version_number = int(time.time())
        self.dom["resource"]["manifest"]["timestamp"] = version_number

    def set_update_priority(self, priority: int):
        """Set priority."""
        self.dom["resource"]["manifest"]["priority"] = priority

    def set_vendor_id(self, vendor_id: bytes):
        """Set vendor id."""
        self.dom["resource"]["manifest"]["vendorId"] = vendor_id

    def set_vendor_data(self, data: bytes):
        """Set vendor data."""
        self.dom["resource"]["manifest"]["vendorInfo"] = data

    def set_class_id(self, class_id: bytes):
        """Set class id."""
        self.dom["resource"]["manifest"]["classId"] = class_id

    def set_device_id(self, device_id: bytes):
        """Set device id."""
        self.dom["resource"]["manifest"]["classId"] = device_id

    def set_payload_fingerprint(self, digest: bytes, size: int):
        """Set payload fingerprint."""
        self.dom["resource"]["manifest"]["payload"]["reference"][
            "hash"
        ] = digest
        self.dom["resource"]["manifest"]["payload"]["reference"]["size"] = size

    def set_payload_uri(self, uri: str):
        """Set payload URI."""
        self.dom["resource"]["manifest"]["payload"]["reference"]["uri"] = uri

    def set_payload_format(self, payload_format: str):
        """Set payload format."""
        _format = payload_format
        if payload_format == "arm-patch-stream":
            _format = "bsdiff-stream"
        self.dom["resource"]["manifest"]["payload"]["format"]["enum"] = _format

    def set_payload_metadata(
        self,
        installed_digest: bytes,
        installed_size: int,
        precursor_digest: bytes,
    ):
        """Set payload metadata."""
        self.dom["resource"]["manifest"]["payload"][
            "installedDigest"
        ] = installed_digest
        self.dom["resource"]["manifest"]["payload"][
            "installedSize"
        ] = installed_size
        self.dom["resource"]["manifest"]["precursorDigest"] = precursor_digest

    def set_update_certificate(self, cert_data: bytes):
        """Set update certificate."""
        hash_ctx = hashlib.sha256()
        hash_ctx.update(cert_data)
        self.fingerprint = hash_ctx.digest()

    def set_image_signature(self, signature: bytes):
        """Set image signature."""

    def set_component_name(self, component: str):
        """Set component name."""



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
# Copyright (c) 2022-2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Setup of manifest tool."""
from setuptools import setup, find_packages, Extension

import manifesttool

armbsdiff = Extension(
    "manifesttool.armbsdiff",
    sources=[
        "bsdiff/bsdiff.c",
        "bsdiff/bsdiff_helper.c",
        "bsdiff/bsdiff_python.c",
        "bsdiff/lz4.c",
        "bsdiff/varint.c",
    ],
    include_dirs=["bsdiff"],
    define_macros=[("LZ4_MEMORY_USAGE", "10")],
    extra_compile_args=["--std=c99", "-O3"],
)

with open("requirements.txt", "rt") as fh:
    tool_requirements = fh.readlines()

setup(
    name="manifest-tool",
    version=manifesttool.__version__,
    description="Tool/lib to create and parse manifests",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/PelionIoT/manifest-tool",
    author="Izuma Networks",
    author_email="opensource@izumanetworks.com",
    license="Apache 2.0",
    packages=find_packages(exclude=["tests"]),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "manifest-tool=manifesttool.mtool.mtool:entry_point",
            "manifest-dev-tool=manifesttool.dev_tool.dev_tool:entry_point",
            "manifest-delta-tool=manifesttool.delta_tool.delta_tool:entry_point",  # noqa: E501
            "manifest-package-tool=manifesttool.package_tool.package_tool:entry_point",  # noqa: E501
        ],
    },
    python_requires=">=3.7.0",
    include_package_data=True,
    install_requires=tool_requirements,
    ext_modules=[armbsdiff],
)



# ----------------------------------------------------------------------------
# Copyright 2019-2021
# Copyright 2022-2023 Izuma Networks
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""Communication with Izuma service."""
import base64
import hashlib
import logging
import time
import urllib.parse
from pathlib import Path
from typing import List, NewType

import requests
import yaml

# APIs documentation:
# https://developer.izumanetworks.com/docs/device-management/current/service-api-references/service-api-documentation.html
# Device directory -
#   https://developer.izumanetworks.com/docs/device-management-api/device-directory
# Update Service -
#   https://developer.izumanetworks.com/docs/device-management-api/update-service

DEVICES = "v3/devices"
DEVICE = "v3/devices/{id}"

FW_IMAGES = "/v3/firmware-images"
FW_IMAGE = "/v3/firmware-images/{id}"
FW_UPLOAD = FW_IMAGES
FW_UPLOAD_JOBS = "/v3/firmware-images/upload-jobs"
FW_UPLOAD_JOB = "/v3/firmware-images/upload-jobs/{id}"
FW_UPLOAD_JOB_CHUNK = "/v3/firmware-images/upload-jobs/{id}/chunks"

FW_MANIFESTS = "/v3/firmware-manifests"
FW_MANIFEST = "/v3/firmware-manifests/{id}"

FW_CAMPAIGNS = "/v3/update-campaigns"
FW_CAMPAIGN = "/v3/update-campaigns/{id}"
FW_CAMPAIGN_STOP = "/v3/update-campaigns/{id}/stop"
FW_CAMPAIGN_START = "/v3/update-campaigns/{id}/start"
FW_CAMPAIGN_STATISTICS = "/v3/update-campaigns/{id}/statistics"
FW_CAMPAIGN_STATISTICS_EVENTS = (
    "/v3/update-campaigns/{id}/statistics/{summary_id}/event_types"
)
FW_CAMPAIGN_DEV_METADATA = "/v3/update-campaigns/{id}/campaign-device-metadata"

CAMPAIGN_ACTIVE_PHASES = ["starting", "active"]

CAMPAIGN_NOT_STARTED_PHASES = ["draft", "stopped"]

FW_UPLOAD_MAX_SMALL_SIZE = int(100 * 1024 * 1024)

# must be greater than 5MB - AWS limitation
FW_UPLOAD_CHUNK_SIZE = int(6 * 1024 * 1024)

LOG = logging.getLogger("izuma")

URL = NewType("URL", str)
ID = NewType("ID", str)
Phase = NewType("Phase", str)

GET_TIMEOUT = 15 * 60
POST_TIMEOUT = 15 * 60
DEL_TIMEOUT = 5 * 60


class IzumaServiceApi:
    """Class for communication with Izuma service."""

    def __init__(self, config_file: Path):
        """
        Create REST API provider for Update Service APIs.

        :param host: Izuma service URL
        :param access_key: account access key
        """
        config = {}
        if config_file.is_file():
            with config_file.open("rt") as fh:
                config = yaml.safe_load(fh)
                if "access_key" not in config and "api_key" in config:
                    # replace api_key with access_key
                    # for backward compatibility
                    config["access_key"] = config.pop("api_key")

        if "host" in config and "access_key" in config:
            self._host = config["host"]
            self._default_headers = {
                "Authorization": "Bearer {}".format(config["access_key"])
            }
            return

        raise AssertionError(
            "Izuma service configurations "
            "(URL and access key) are not provided"
        )

    def _url(self, api, **kwargs) -> str:
        """
        Construct full REST API URL.

        Concatenates Izuma host URL with the desired REST API url and expands
        any patterns present in the URL (e.g. v3/manifests/{id})
        :param api: REST API URL part
        :param kwargs: dictionary for expanding the the URL patterns
        :return: full URL
        """
        if kwargs:
            return urllib.parse.urljoin(self._host, api.format(**kwargs))
        return urllib.parse.urljoin(self._host, api)

    def _headers(self, extra_headers: dict = None) -> dict:
        """
        Create request header.

        Initializes a dictionary with common headers and allows extending it
        with REST API specific data
        :param extra_headers: dictionary containing extra headers to be sent as
        part of REST API request
        :return: headers dictionary
        """
        copy = self._default_headers.copy()
        if extra_headers:
            copy.update(extra_headers)
        return copy

    @staticmethod
    def _chunk_md5(chunk: bytes) -> str:
        """
        Calculate MD5 digest over bytes.

        :param chunk: bytes to be processed
        :return: base64 encoded MD5 digest
        """
        return base64.b64encode(hashlib.md5(chunk).digest()).decode()  # nosec

    @staticmethod
    def _print_upload_progress(current: int, total: int):
        """
        Print upload progress.

        :param current: bytes uploaded so far
        :param total: uploaded file size
        """
        if total < FW_UPLOAD_CHUNK_SIZE:
            return
        increments = 50
        progress = (current / total) * 100
        count = int(progress // (100 / increments))
        text = "\r[{0: <{1}}] {2:.2f}%".format(
            "=" * count, increments, progress
        )
        print(text, end="\n" if progress == 100 else "")

    def fw_upload(self, fw_name: str, image: Path, encrypt: bool) -> dict:
        """
        Upload FW image.

        :param fw_name: update candidate image name as will appear on
               Izuma portal
        :param image: candidate FW image
        :param encrypt: request to encrypt FW image
        :return: uploaded image meta
        """
        fw_size = image.stat().st_size

        if fw_size < FW_UPLOAD_MAX_SMALL_SIZE:
            fw_meta = self._upload_small_image(fw_name, image, encrypt)
        else:
            fw_meta = self._upload_big_image(fw_name, fw_size, image, encrypt)

        LOG.info("Uploaded FW image %s", fw_meta["datafile"])
        return fw_meta

    def _upload_image_chunk(self, chunk: bytes, job_id: ID):
        """
        Upload FW image chunk.

        :param chunk: FW image chunk to be uploaded
        :param job_id: upload job ID
        """
        response = requests.post(
            self._url(FW_UPLOAD_JOB_CHUNK, id=job_id),
            headers=self._headers(
                {
                    "Content-Type": "binary/octet-stream",
                    "Content-MD5": self._chunk_md5(chunk),
                }
            ),
            data=chunk,
            timeout=POST_TIMEOUT,
        )
        response.raise_for_status()

    def _upload_big_image(
        self, fw_name: str, fw_size: int, image: Path, encrypt: bool
    ) -> dict:
        """
        Upload small FW image.

        :param fw_name: fw name as appears on the portal
        :param image: candidate FW image
        :param encrypt: Request to encrypt FW image
        :return: uploaded image meta
        """
        job_id = None
        try:
            with image.open("rb") as fh:
                job_id = self._create_upload_job(fw_name, encrypt)
                self._print_upload_progress(0, fw_size)
                upload_counter = 0
                while True:
                    chunk = fh.read(FW_UPLOAD_CHUNK_SIZE)
                    upload_counter += len(chunk)
                    self._upload_image_chunk(chunk, job_id)
                    self._print_upload_progress(upload_counter, fw_size)
                    if not chunk:
                        break
            LOG.debug("FW upload job %s completed", job_id)
            return self._get_fw_image_meta(job_id)
        except requests.HTTPError:
            LOG.error("FW image upload failed")
            raise
        finally:
            if job_id:
                self._delete_upload_job(job_id)

    def _upload_small_image(
        self, fw_name: str, image: Path, encrypt: bool
    ) -> dict:
        """
        Upload small FW image.

        :param fw_name: fw name as appears on the portal
        :param image: candidate FW image
        :param encrypt: Request to encrypt FW image
        :return: uploaded image meta
        """
        with image.open("rb") as fh:
            response = requests.post(
                self._url(FW_UPLOAD),
                headers=self._headers(),
                files={
                    "datafile": (image.name, fh),
                },
                data={"name": fw_name, "datafile_encryption": encrypt},
                timeout=POST_TIMEOUT,
            )
            response.raise_for_status()
            return response.json()

    def _get_fw_image_meta(self, job_id: ID) -> dict:
        """
        Extract uploaded image URL and ID.

        :param job_id: upload job ID
        :return: uploaded image meta
        """
        response = requests.get(
            self._url(FW_UPLOAD_JOB, id=job_id),
            headers=self._headers(),
            timeout=GET_TIMEOUT,
        )
        response.raise_for_status()
        image_id = response.json()["firmware_image_id"]
        response = requests.get(
            self._url(FW_IMAGE, id=image_id),
            headers=self._headers(),
            timeout=GET_TIMEOUT,
        )
        response.raise_for_status()

        return response.json()

    def _delete_upload_job(self, job_id: ID):
        """
        Delete a firmware image upload job.

        :param job_id: upload job ID
        """
        response = requests.delete(
            self._url(FW_UPLOAD_JOB, id=job_id),
            headers=self._headers(),
            timeout=DEL_TIMEOUT,
        )
        response.raise_for_status()
        LOG.debug("FW upload job %s deleted", job_id)

    def _create_upload_job(self, fw_name: str, encrypt: bool) -> ID:
        """
        Create a firmware image upload job.

        :param fw_name: fw name as appears on the portal
        :param encrypt: Request to encrypt FW image
        :return: upload job ID
        """
        response = requests.post(
            self._url(FW_UPLOAD_JOBS),
            headers=self._headers(),
            json={"name": fw_name, "datafile_encryption": encrypt},
            timeout=POST_TIMEOUT,
        )
        response.raise_for_status()
        response_data = response.json()
        job_id = response_data["id"]
        LOG.debug("FW upload job created %s", job_id)
        return job_id

    def fw_delete(self, image_id: ID):
        """
        Delete candidate image from Izuma portal.

        :param image_id: image ID as appears on the portal
        """
        response = requests.delete(
            self._url(FW_IMAGE, id=image_id),
            headers=self._headers(),
            timeout=DEL_TIMEOUT,
        )
        response.raise_for_status()
        LOG.info("Deleted FW image %s", image_id)

    def manifest_upload(self, name: str, manifest: Path) -> ID:
        """
        Upload manifest file to Izuma service.

        :param name: manifest name as will appear on Izuma portal
        :param manifest: manifest file
        :return: manifest ID as reported by portal
        """
        try:
            with manifest.open("rb") as fh:
                response = requests.post(
                    self._url(FW_MANIFESTS),
                    headers=self._headers(),
                    files={
                        "datafile": (manifest.name, fh),
                    },
                    data={"name": name},
                    timeout=POST_TIMEOUT,
                )
                response.raise_for_status()
                manifest_id = response.json()["id"]
                LOG.info("Uploaded manifest ID: %s", manifest_id)
                return manifest_id
        except requests.HTTPError:
            LOG.error("Failed to upload manifest")
            raise

    def manifest_delete(self, manifest_id: ID):
        """
        Delete manifest file from Izuma portal.

        :param manifest_id: manifest ID to be deleted
        """
        response = requests.delete(
            self._url(FW_MANIFEST, id=manifest_id),
            headers=self._headers(),
            timeout=DEL_TIMEOUT,
        )
        response.raise_for_status()
        LOG.info("Deleted manifest ID: %s", manifest_id)

    def campaign_create(
        self, name: str, manifest_id: ID, device_filter: str
    ) -> ID:
        """
        Create update campaign on Izuma portal.

        :param name: campaign name as will appear on Izuma portal
        :param manifest_id: manifest ID as
        :param device_filter: device filter query
        :return: campaign ID as reported by portal
        """
        try:
            response = requests.post(
                self._url(FW_CAMPAIGNS),
                headers=self._headers(),
                json={
                    "campaign_strategy": "one-shot",
                    "description": "Development campaign",
                    "device_filter": device_filter,
                    "name": name,
                    "root_manifest_id": manifest_id,
                },
                timeout=POST_TIMEOUT,
            )
            response.raise_for_status()
            campaign_id = response.json()["id"]
            LOG.info("Created Campaign ID: %s", campaign_id)
            return campaign_id
        except requests.HTTPError:
            LOG.error("Failed to create campaign")
            raise

    def campaign_delete(self, campaign_id: ID):
        """
        Delete inactive (stopped/draft) update campaign.

        :param campaign_id: campaign ID
        """
        response = requests.delete(
            self._url(FW_CAMPAIGN, id=campaign_id),
            headers=self._headers(),
            timeout=DEL_TIMEOUT,
        )
        response.raise_for_status()
        LOG.info("Deleted campaign %s", campaign_id)

    def campaign_stop(self, campaign_id: ID, timeout: int = 60):
        """
        Stop update campaign.

        :param campaign_id: campaign ID
        :param timeout: timeout in seconds to wait for campaign to stop
        """
        try:
            # check campaign phase and skip if it not active
            curr_phase = self.campaign_get(campaign_id)["phase"]
            if not self.campaign_is_active(curr_phase):
                return
            # send request to stop
            response = requests.post(
                self._url(FW_CAMPAIGN_STOP, id=campaign_id),
                headers=self._headers(),
                timeout=POST_TIMEOUT,
            )
            response.raise_for_status()
            curr_phase = self.campaign_get(campaign_id)["phase"]
            # The campaign is stopping. wait for it to stop
            retries = timeout
            while curr_phase == "stopping" and retries > 0:
                time.sleep(1)
                curr_phase = self.campaign_get(campaign_id)["phase"]
                retries -= 1
            if retries == 0:
                LOG.debug("Stopping Campaign timed out")
            LOG.info("Stopped campaign %s", campaign_id)
        except requests.HTTPError as ex:
            # ignore error from stop campaign as it may already be stopped
            LOG.debug("Failed stopping campaign - %s", ex)

    def campaign_start(self, campaign_id: ID):
        """
        Start update campaign.

        :param campaign_id: campaign ID
        """
        response = requests.post(
            self._url(FW_CAMPAIGN_START, id=campaign_id),
            headers=self._headers(),
            timeout=POST_TIMEOUT,
        )
        response.raise_for_status()

    def campaign_get(self, campaign_id: ID) -> dict:
        """
        Get an update campaign current representation/state.

        :param campaign_id:
        :return: Dictionary with current update campaign state
        """
        response = requests.get(
            self._url(FW_CAMPAIGN, id=campaign_id),
            headers=self._headers(),
            timeout=GET_TIMEOUT,
        )
        response.raise_for_status()
        return response.json()

    def campaign_statistics(self, campaign_id: ID) -> List[dict]:
        """
        Get campaign statistics.

        :param campaign_id: campaign ID
        :return: List of statistics for a campaign
        """
        try:
            response = requests.get(
                self._url(FW_CAMPAIGN_STATISTICS, id=campaign_id),
                headers=self._headers(),
                timeout=GET_TIMEOUT,
            )
            response.raise_for_status()
            return response.json()["data"]
        except requests.HTTPError:
            pass
        return []

    def campaign_statistic_events(
        self, campaign_id: ID, summary_id: str
    ) -> List[dict]:
        """
        Get campaign events grouped by summary.

        :param campaign_id: campaign ID
        :param summary_id: The summary status
            Available values: fail, success, info, skipped
        :return: List of statistics for a campaign
        """
        try:
            response = requests.get(
                self._url(
                    FW_CAMPAIGN_STATISTICS_EVENTS,
                    id=campaign_id,
                    summary_id=summary_id,
                ),
                headers=self._headers(),
                timeout=GET_TIMEOUT,
            )
            response.raise_for_status()
            return response.json()["data"]
        except requests.HTTPError:
            pass
        return []

    def campaign_device_metadata(self, campaign_id: ID) -> List[dict]:
        """
        Get metadata for devices participating in update campaign.

        Note: this function and assumes small number of devices participating
        in update campaign (as part of developer flow). Thus it does not
        handles pagination
        :param campaign_id: campaign ID
        :return: List of device metadata dictionaries for devices participating
        in update campaign
        """
        response = requests.get(
            self._url(FW_CAMPAIGN_DEV_METADATA, id=campaign_id),
            headers=self._headers(),
            timeout=GET_TIMEOUT,
        )
        response.raise_for_status()
        return response.json()["data"]

    @staticmethod
    def campaign_is_active(phase: Phase) -> bool:
        """
        Understand if update campaign is in active phase.

        :param phase: current update campaign phase
        :return: True if update campaign phase is starting or active
        """
        return phase in CAMPAIGN_ACTIVE_PHASES

    @staticmethod
    def campaign_is_not_started(phase: Phase) -> bool:
        """
        Understand if update campaign is not started phase.

        :param phase: current update campaign phase
        :return: True if update campaign phase is starting or active
        """
        return phase in CAMPAIGN_NOT_STARTED_PHASES

    def device_delete(self, device_id: ID):
        """
        Delete device with a developer certificate.

        :param device_id: device ID
        """
        response = requests.delete(
            self._url(DEVICE, id=device_id),
            headers=self._headers(),
            timeout=DEL_TIMEOUT,
        )
        response.raise_for_status()
        LOG.info("Deleted device %s", device_id)

    def _get_objects(self, api_url: str, api_filter: str = "") -> dict:
        """
        Get list of objects for the account.

        :param api_filter: filter query string
        :return: Dictionary with the list of objects
        """
        api_filter = "&filter={}".format(api_filter) if api_filter else ""
        url = "{}?limit=1000&include=total_count{}".format(api_url, api_filter)
        response = requests.get(
            self._url(url), headers=self._headers(), timeout=GET_TIMEOUT
        )
        response.raise_for_status()
        total_count = response.json()["total_count"]
        objects = response.json()["data"]
        LOG.debug("Got %d/%d objects", len(objects), total_count)

        url = url + "&after={after}"
        while len(objects) < total_count:
            last_object_id = objects[-1]["id"]
            response = requests.get(
                self._url(url, after=last_object_id),
                headers=self._headers(),
                timeout=GET_TIMEOUT,
            )
            response.raise_for_status()
            objects.extend(response.json()["data"])
            LOG.debug("Got %d/%d objects", len(objects), total_count)

        return objects

    def get_devices(self, api_filter: str = "") -> dict:
        """
        Get list of devices for the account.

        :param api_filter: filter query string
        :return: Dictionary with the list of devices
        """
        return self._get_objects(DEVICES, api_filter)

    def get_fw_images(self, api_filter: str = "") -> dict:
        """
        Get list of firmware images for the account.

        :param api_filter: filter query string
        :return: Dictionary with the list of firmware images
        """
        return self._get_objects(FW_IMAGES, api_filter)

    def get_manifests(self, api_filter: str = "") -> dict:
        """
        Get list of manifests for the account.

        :param api_filter: filter query string
        :return: Dictionary with the list of manifests
        """
        return self._get_objects(FW_MANIFESTS, api_filter)

    def get_campaigns(self, api_filter: str = "") -> dict:
        """
        Get list of campaigns for the account.

        :param api_filter: filter query string
        :return: Dictionary with the list of campaigns
        """
        return self._get_objects(FW_CAMPAIGNS, api_filter)
